/* $Id: safespawn.c,v 1.2 2000/10/10 22:10:44 dps Exp $ */
/* Front end that spawns a process using fork() and safe_exec() */

#define _POSIX_SOURCE 1

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif /* HAVE_CONFIG_H */
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif /* HAVE_SYS_TYPES_H */
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif /* HAVE_SYS_WAIT_H */
#ifdef HAVE_SETJMP_H
#include <setjmp.h>
#endif /* HAVE_SETJMP_H */
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif /* HAVE_ERRNO_H */
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif /* HAVE_SIGNAL_H */

/* update RW Tue Dec 30 15:41:04 CET 2003 */
#define __need_timespec 1
#include <time.h>
/* update end */

#include "../prog.h"
#include "detail.h"


/* Use longjmp and setjmp if siglongjmp and sigsetjmp are not
   avialable with macro magic */
#ifndef HAVE_SIGLONGJMP
#define siglongjmp(x) longjmp(x)
#endif /* HAVE_SIGLONGJMP */
#ifndef HAVE_SIGSETJMP
#define sigsetjmp(x,y) setjmp(x)
#endif /* HAVE_SIGSETJMP */

/* Use signal or sigaction to restore a signal */
#ifdef HAVE_SIGACTION
#define restore_sig(x,y) sigaction(x, &y, NULL)
#else
#define restore_sig(x,y) signal(x,y)
#endif /* HAVE_SIGACTION */


/* Globals */
#ifdef HAVE_SIGACTION
static struct sigaction oldalrm, oldchld;
#else
static __sighandler_t oldalrm, oldchld;
#endif /* HAVE_SIGACTION */
static const char *child_name="<unset>";
#ifdef HAVE_SIGSETJMP
static sigjmp_buf timeout_env;
#else
static jmp_buf timeout_env;
#endif /* HAVE_SIGSETJMP */
static pid_t pid;

/* Return pid */
pid_t spawned_pid(void)
{
    return pid;
}

/* Alarm signal handler */
static RETSIGTYPE timeout(int sig)
{
    sig=sig;
    siglongjmp(timeout_env,1);
    /* Not reached */
}

/* Child catcher */
static RETSIGTYPE catch_child(int sig)
{
    int status;
#ifndef HAVE_SIGACTION
    __sighandler_t alrm;
    alrm=signal(SIGALRM, SIG_IGN);	/* Squash alarms temporalily */
#endif /* HAVE_SIGACTION */

    pid=wait(&status);
    if (WIFSTOPPED(status))
    {
	do_log("child %s unsuspended (stopped by sig %d)", child_name,
	    WSTOPSIG(status));
	kill(pid, SIGCONT);		/* Stop crackers stoping theprocess */
					/* (major paranoia) */
	signal(SIGCHLD, catch_child);	/* Reattach myself */
#ifndef HAVE_SIGACTION
	signal(SIGALRM, alrm);		/* Restore alarms */
#endif /* HAVE_SIGACTION */
	RETSIG;				/* alais return; or return 0; */
    }
    else if (WIFEXITED(status))
    {
	if (WEXITSTATUS(status)!=0)
	    do_log("%s exited with status %d", child_name, WEXITSTATUS(status));
    }
    else if (WIFSIGNALED(status))
    {
	do_log("%s terminated with signal %d", child_name,  WTERMSIG(status));
    }
    restore_sig(SIGCHLD, oldchld);	/* Restore old SIGCHLD handler */
    alarm(0);				/* Turn off alarm */
    restore_sig(SIGALRM, oldalrm);	/* Restore old SIGALRM handler */
    RETSIG;				/* alais return; or return 0; */
}


FILE *safe_spawn(const char *const path, const char *const *args, unsigned int maxtime)
{
    int fd[2];
    struct stat stbuf;
    FILE *in;
#ifdef HAVE_SIGACTION
    struct sigaction chld, alrm;
#endif /* HAVE_SIGACTION */

    /* Stat in the parent */
    if ((lstat(path, &stbuf))!=0)
    {
	do_log("Could not stat %s (error %s)", path, strerror(errno));
	return NULL;
    }
    /* Add not LRU killable cache entry */
    inode_cache(path, stbuf.st_ino, stbuf.st_dev, stbuf.st_mtime,
		INODE_KEEP | INODE_MTIME); 

    /* Create a pipe for the output */
    if (pipe(fd))
    {
       do_log("Could not create pipe");
       return NULL;
    }

    if ((in=fdopen(fd[0],"r"))==NULL)
    {
       do_log("fdopen failure");
       close(fd[0]);
       close(fd[1]);
       return NULL;
    }

    /* Set up signal handlers */
    child_name=*args;
#ifdef HAVE_SIGACTION
    chld.sa_handler=catch_child;
    sigemptyset(&(chld.sa_mask));
    sigaddset(&(chld.sa_mask), SIGALRM);
    chld.sa_flags=0;
    sigaction(SIGCHLD, &chld, &oldchld);

    alrm.sa_handler=timeout;
    sigemptyset(&(alrm.sa_mask));
    sigaddset(&(alrm.sa_mask), SIGCHLD);
    alrm.sa_flags=0;
    sigaction(SIGALRM, &alrm, &oldalrm);
#else
    oldchld=signal(SIGCHLD, catch_child);
    oldalrm=signal(SIGALRM, timeout);
#endif /* HAVE_SIGACTION */
    pid=fork();			     /* Need right value of pid at setjmp */

    if (pid!=0)
    {
	/* Parent */
	if (sigsetjmp(timeout_env, 1)!=0)
	{
	    if (pid!=-1)
		kill(pid, SIGKILL);	     /* Kill ps */
	    alarm(0);			     /* Disable alarm */
	    restore_sig(SIGALRM, oldalrm);   /* Reset signal handling */
	    do_log("%s output timeout", *args); /* Nasty program */
	    log_flush();		     /* Tell admin */
	    /* Child handler will restore old SIGCHLD handler for us */
	    return 0;			     /* Failed */
	}
#ifndef TEST
	alarm(maxtime);		     /* Set alarm */
#endif /* TEST */
    }
    switch(pid)
    {
    case -1:
       do_log("Fork failed with error %s", strerror(errno));
       close(fd[0]);
       close(fd[1]);
       alarm(0);			     /* Turn off alarm */
       restore_sig(SIGALRM, oldalrm);	     /* Reset signal handling */
       restore_sig(SIGCHLD, oldchld);	     /* Restore SIGCHLD */
       return NULL;

    case 0: /* child */
	close(fd[0]);
	restore_sig(SIGALRM, oldalrm);	     /* Reset signal handling */
	restore_sig(SIGCHLD, oldchld);	     /* Restore SIGCHLD */
	safe_exec(path, args, fd[1]);
	/* NOT REACHED */
    default:
       close(fd[1]);
       /* update RW Tue Dec 30 15:41:04 CET 2003 */
       {
	 /* let the child exit to avoid 'broken pipe' */
	 struct timespec req, rem;
	 req.tv_sec  = 0;
	 req.tv_nsec = 1000000;
	 nanosleep(&req, &rem);
       }
       /* update end */
       break;
    }
    return in;
}
